# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
from urllib.parse import urlparse

from contrast import AGENT_CURR_WORKING_DIR

from contrast.agent.framework import Framework, Server
from contrast.agent.protect.mixins.REP_settings import SettingsREPMixin
from contrast.agent.reaction_processor import ReactionProcessor
from contrast.agent.protect.rule import ProtectionRule
from contrast.api.settings_pb2 import (
    AccumulatorSettings as DtmAccumulatorSettings,
    AgentSettings as DtmAgentSettings,
    ApplicationSettings as DtmApplicationSettings,
    ServerFeatures as DtmServerFeatures,
)
from contrast.configuration import AgentConfig, SamplingLockConfig
from contrast.utils.decorators import cached_property
from contrast.utils.loggers.logger import reset_agent_logger
from contrast.utils.singleton import Singleton
from contrast.utils.string_utils import truncate
from contrast.utils.timer import now_ms
from contrast.extern import structlog as logging
from contrast.agent.exclusions import Exclusions, ExclusionMapper

logger = logging.getLogger("contrast")

ASSESS_STACKTRACES = "assess.stacktraces"


class ServerFeatures:
    def __init__(self, features):
        self.features = features or {}

    @cached_property
    def log_file(self):
        return self.features.get("features", {}).get("logFile", "")

    @cached_property
    def log_level(self):
        return self.features.get("features", {}).get("logLevel", "")

    @cached_property
    def assess_enabled(self):
        return (
            self.features.get("features", {})
            .get("assessment", {})
            .get("enabled", False)
        )

    @cached_property
    def protect_enabled(self):
        return self.features.get("features", {}).get("defend", {}).get("enabled", False)


class ApplicationSettings:
    def __init__(self, settings_json):
        self._raw_settings = settings_json or {}

    @cached_property
    def disabled_assess_rules(self):
        return (
            self._raw_settings.get("settings", {})
            .get("assessment", {})
            .get("disabledRules", [])
        )

    @cached_property
    def protection_rules(self):
        return [
            ProtectRule(r)
            for r in self._raw_settings.get("settings", {})
            .get("defend", {})
            .get("protectionRules", [])
        ]

    @cached_property
    def session_id(self):
        return (
            self._raw_settings.get("settings", {})
            .get("assessment", {})
            .get("session_id")
        )

    @cached_property
    def sensitive_data_masking_policy(self):
        return self._raw_settings.get("settings", {}).get(
            "sensitive_data_masking_policy"
        )

    @cached_property
    def exclusions(self):
        return self._raw_settings.get("settings", {}).get("exceptions")

    # note: there are many more fields on the ApplicationSettings that we currently do
    # not use


class ProtectRule:
    def __init__(self, protect_rule_json):
        self._raw_rule = protect_rule_json or {}

    @cached_property
    def id(self):
        return self._raw_rule.get("id", "")

    @cached_property
    def mode(self):
        raw_mode = self._raw_rule.get("mode", "OFF")
        raw_block_at_entry = self._raw_rule.get("blockAtEntry", False)
        return {
            "MONITORING": ProtectionRule.MONITOR,
            "BLOCKING": (
                ProtectionRule.BLOCK_AT_PERIMETER
                if raw_block_at_entry
                else ProtectionRule.BLOCK
            ),
            "OFF": ProtectionRule.NO_ACTION,
        }[raw_mode]


class Settings(Singleton, SettingsREPMixin):
    def init(self, app_name=None, framework_name=None):
        """
        Agent settings for the entire lifetime of the agent.

        Singletons should override init, not __init__.
        """
        self.config = None
        self.config_features = {}
        self.exclusions = None
        self.last_update_service = None
        self.last_server_update_time_ms = 0
        self.last_app_update_time_ms = 0
        self.heartbeat = None
        self.server_settings_poll = None
        self.framework = Framework(framework_name)
        self.server = Server()
        self.sys_module_count = 0
        self.rewriter_enabled = False

        # Features and Settings from Service
        self.server_features = DtmServerFeatures()

        self.application_settings = DtmApplicationSettings()
        self.accumulator_settings = DtmAccumulatorSettings()

        # Server
        self.server_name = None
        self.server_path = None
        self.server_type = None

        self.exclusion_matchers = []

        # Rules
        self.protect_rules = dict()

        # circular import
        from contrast.agent.assess.rules.response.autocomplete_missing_rule import (
            AutocompleteMissingRule,
        )
        from contrast.agent.assess.rules.response.cache_controls_rule import (
            CacheControlsRule,
        )
        from contrast.agent.assess.rules.response.clickjacking_rule import (
            ClickjackingRule,
        )

        from contrast.agent.assess.rules.response.x_content_type_rule import (
            XContentTypeRule,
        )

        from contrast.agent.assess.rules.response.csp_header_missing_rule import (
            CspHeaderMissingRule,
        )

        from contrast.agent.assess.rules.response.x_xss_protection_disabled_rule import (
            XXssProtectionDisabledRule,
        )

        from contrast.agent.assess.rules.response.hsts_header_rule import HstsHeaderRule

        from contrast.agent.assess.rules.response.csp_header_insecure_rule import (
            CspHeaderInsecureRule,
        )

        from contrast.agent.assess.rules.response.parameter_pollution_rule import (
            ParameterPollutionRule,
        )

        self.assess_response_rules = [
            AutocompleteMissingRule(),
            CacheControlsRule(),
            ClickjackingRule(),
            XContentTypeRule(),
            CspHeaderMissingRule(),
            XXssProtectionDisabledRule(),
            HstsHeaderRule(),
            CspHeaderInsecureRule(),
            ParameterPollutionRule(),
        ]

        # Initialize config
        self.config = AgentConfig()

        if self.config.is_service_bypassed:
            self.direct_teamserver_server_features = ServerFeatures(None)
            self.direct_ts_app_settings = ApplicationSettings(None)

        self.disabled_assess_rules = set(
            self.config.get("assess.rules.disabled_rules", [])
        )

        # Initialize application metadata
        self.app_name = self.get_app_name(app_name)

        self.agent_runtime_window = now_ms()

        logger.info("Contrast Agent finished loading settings.")

    def _is_defend_enabled_in_server_features(self):
        if self.config.is_service_bypassed:
            return self.direct_teamserver_server_features.protect_enabled
        return (
            self.server_features
            and self.server_features.defend
            and self.server_features.defend.enabled
        )

    @cached_property
    def is_proxy_enabled(self):
        return self.config.get("api.proxy.enable")

    @cached_property
    def proxy_url(self):
        return self.config.get("api.proxy.url")

    @cached_property
    def proxy_scheme(self):
        scheme = urlparse(self.proxy_url).scheme or "http"
        return scheme

    @cached_property
    def is_cert_verification_enabled(self):
        return self.config.get("api.certificate.enable")

    @cached_property
    def ca_file(self):
        return self.config.get("api.certificate.ca_file")

    @cached_property
    def client_cert_file(self):
        return self.config.get("api.certificate.cert_file")

    @cached_property
    def client_private_key(self):
        return self.config.get("api.certificate.key_file")

    @cached_property
    def api_service_key(self):
        return self.config.get("api.service_key")

    @cached_property
    def api_url(self):
        return self.config.get("api.url")

    @cached_property
    def api_key(self):
        return self.config.get("api.api_key")

    @cached_property
    def api_user_name(self):
        return self.config.get("api.user_name")

    def is_agent_config_enabled(self):
        return self.config.get("enable", True)

    @cached_property
    def is_rewriter_enabled(self):
        return self.config.get("agent.python.rewrite")

    @cached_property
    def is_profiler_enabled(self):
        return self.config.get("agent.python.enable_profiler")

    @cached_property
    def max_sources(self):
        return self.config.get("assess.max_context_source_events")

    @cached_property
    def max_propagation(self):
        return self.config.get("assess.max_propagation_events")

    @cached_property
    def max_vulnerability_count(self):
        """Max number of vulnerabilities per rule type to report for one
        agent run `time_limit_threshold` time period"""
        return self.config.get("assess.max_rule_reported")

    @cached_property
    def agent_runtime_threshold(self):
        return self.config.get("assess.time_limit_threshold")

    @cached_property
    def is_bundled_enabled(self):
        """The default across agents is to use the bundled service by default"""
        return self.config.get("agent.service.enable", True)

    @cached_property
    def app_path(self):
        return self.config.get("application.path")

    @cached_property
    def app_version(self):
        return self.config.get("application.version")

    @property
    def pid(self):
        """
        pid is used by Speedracer to recognize a unique worker process for an application.

        pid must be unique for each worker process of an app.
        :return: int current process id
        """
        return os.getpid()

    def get_app_name(self, app_name):
        if self.config.get("application.name"):
            return self.config.get("application.name")

        return app_name if app_name else "root"

    def establish_heartbeat(self):
        """
        Initialize Heartbeat between Agent and SR if it has not been already initialized.
        """
        if self.heartbeat is None:
            # Circular import
            from contrast.agent.heartbeat import Heartbeat

            self.heartbeat = Heartbeat()
            self.heartbeat.start()

    def establish_server_settings_poll(self):
        """
        Initialize Server Settings poll between Agent and TS if it has not been already initialized.
        """
        if self.server_settings_poll is None:
            # Circular import
            from contrast.agent.server_settings_poll import ServerSettingsPoll

            self.server_settings_poll = ServerSettingsPoll()
            self.server_settings_poll.start()

    def process_responses(self, responses):
        """
        :param responses: list of Message responses from SR
        """
        logger.debug("Processing %s responses", len(responses))

        for response in responses:
            if self.process_service_response(response):
                self.set_protect_rules()

    def process_service_response(self, data):
        reload_rules = False

        if data and isinstance(data, DtmAgentSettings):
            self.last_update_service = data.sent_ms

            reload_rules = self.process_server_features(data) or reload_rules
            reload_rules = self.process_application_settings(data) or reload_rules
            reload_rules = self.process_accumulator_settings(data) or reload_rules

        if reload_rules:
            logger.debug(
                "Finished processing Contrast Service message and reloading rules."
            )

        return reload_rules

    def apply_ts_server_settings(self, response_body):
        self.direct_teamserver_server_features = ServerFeatures(response_body)

        if (
            self.direct_teamserver_server_features.features.get("features", {})
            .get("assessment", {})
            .get("sampling")
        ):
            self.update_sampling()

        self.update_logger_from_features()

        self.log_server_features(self.direct_teamserver_server_features)

        self.set_protect_rules()

        self.last_server_update_time_ms = now_ms()

    def process_ts_reactions(self, response_body):
        # App startup/activity wrap reactions in a settings dict whereas
        # Server startup/activity has it at the top level response dict
        reactions = response_body.get("settings", {}).get("reactions", None)

        if not reactions:
            reactions = response_body.get("reactions", None)

        if not reactions:
            return

        ReactionProcessor.process(reactions, self)

    def process_server_features(self, data):
        if not data.HasField("server_features"):
            return False

        self.server_features = data.server_features
        self.update_logger_from_features()

        if not self.config.is_service_bypassed and self.server_features.assess.enabled:
            self.update_sampling()

        self.log_server_features(data.server_features)

        return True

    def log_server_features(self, server_features):
        """
        Record server features received from teamserver (via the contrast service)
        Remove the rule_definitions field before logging because it's long and ugly
        """
        if isinstance(server_features, ServerFeatures):
            logger.debug(
                "Received updated server features logFile=%s logLevel=%s Protect=%s Assess=%s",
                server_features.log_file,
                server_features.log_level,
                server_features.protect_enabled,
                server_features.assess_enabled,
            )
        else:
            server_features_copy = DtmServerFeatures()
            server_features_copy.CopyFrom(server_features)
            del server_features_copy.defend.rule_definitions[:]
            logger.debug(
                "Received updated server features (excluding rule_definitions from"
                " log):\n%s",
                server_features_copy,
            )

    @property
    def code_exclusion_matchers(self):
        return [x for x in self.exclusion_matchers if x.is_code]

    def load_exclusions_from_server(self, application_settings):
        if (
            hasattr(application_settings, "exclusions")
            and application_settings.exclusions
        ):
            self.exclusions = Exclusions(
                application_settings.exclusions,
                self.is_protect_enabled(),
                self.is_assess_enabled(),
            )

    def evaluate_exclusions(self, context, path):
        if self.exclusions:
            if self.exclusions.evaluate_url_exclusions(context, path):
                # Stop analyzing this endpoint since the URL exclusion applies
                return True

            if self.exclusions.input_exclusions:
                self.exclusions.set_input_exclusions_by_url(context, path)

        return False

    def apply_ts_app_settings(self, response_json):
        self.direct_ts_app_settings = ApplicationSettings(response_json)

        if self.direct_ts_app_settings.exclusions:
            em = ExclusionMapper(self.direct_ts_app_settings.exclusions)

            self.load_exclusions_from_server(em)

        # This is the only place session_id is set by TS.
        # If session id is set in the config, that value will be echoed back by TS
        if self.direct_ts_app_settings.session_id:
            self.config.session_id = self.direct_ts_app_settings.session_id

    def process_application_settings(self, data):
        if not data.HasField("application_settings"):
            return False

        self.application_settings = data.application_settings

        if self.application_settings.session_id:
            self.config.session_id = self.application_settings.session_id

        ReactionProcessor.process(data.application_settings, self)

        if not self.config.is_service_bypassed:
            self.load_exclusions_from_server(self.application_settings)

        self.reset_transformed_settings()

        logger.debug(
            "Received updated application settings:\n%s", data.application_settings
        )

        return True

    def reset_transformed_settings(self):
        self.exclusion_matchers = []

    def process_accumulator_settings(self, data):
        if data.HasField("accumulator_settings"):
            self.accumulator_settings = data.accumulator_settings

    def update_logger_from_features(self):
        if self.config.is_service_bypassed:
            logger_reset = reset_agent_logger(
                self.direct_teamserver_server_features.log_file,
                self.direct_teamserver_server_features.log_level,
            )
        else:
            logger_reset = reset_agent_logger(
                self.server_features.log_file, self.server_features.log_level
            )

        if logger_reset:
            self.config.log_config()

    def is_inventory_enabled(self):
        """
        inventory.enable = false: Disables both route coverage and library analysis and reporting
        """
        return self.config.get("inventory.enable", True)

    def is_analyze_libs_enabled(self):
        """
        inventory.analyze_libraries = false: Disables only library analysis/reporting
        """
        return (
            self.config is not None
            and self.config.get("inventory.analyze_libraries", True)
            and self.is_inventory_enabled()
        )

    def is_assess_enabled(self):
        """
        We do not allow assess and defend to be on at the same time. As defend
        is arguably the more important of the two, it will take precedence

        The agent config may enable assess even if it is turned off in TS. This
        allows unlicensed apps to send findings to TS, where they will appear
        as obfuscated results.
        """

        # https://contrast.atlassian.net/browse/PROD-530
        if self.config is None:
            return False

        assess_enabled = self.config.get("assess.enable", None)
        if assess_enabled is not None:
            return assess_enabled

        if self.config.is_service_bypassed:
            return self.direct_teamserver_server_features.assess_enabled
        return (
            self.server_features
            and self.server_features.assess
            and self.server_features.assess.enabled
        )

    def is_protect_enabled(self):
        """
        Protect is enabled only if both configuration and server features enable it.
        """
        if self.config is None:
            return False
        config_protect_enabled = self.config.get("protect.enable", None)
        if config_protect_enabled is not None:
            return config_protect_enabled

        return self._is_defend_enabled_in_server_features()

    def set_protect_rules(self):
        if not self._is_defend_enabled_in_server_features():
            self.protect_rules = dict()
            return

        from contrast.agent.protect.rule.rules_builder import build_protect_rules

        self.protect_rules = build_protect_rules()

    def get_server_name(self):
        """
        Hostname of the server

        Default is socket.gethostname() or localhost
        """
        if self.server_name is None:
            self.server_name = self.config.get("server.name")

        return self.server_name

    def get_server_path(self):
        """
        Working Directory of the server

        Default is root
        """
        if self.server_path is None:
            self.server_path = self.config.get("server.path") or truncate(
                AGENT_CURR_WORKING_DIR
            )

        return self.server_path

    def get_server_type(self):
        """
        Web Framework of the Application either defined in common config or via discovery.
        """
        if self.server_type is None:
            self.server_type = (
                self.config.get("server.type") or self.framework.name_lower
            )

        return self.server_type

    @property
    def response_scanning_enabled(self):
        return self.is_assess_enabled() and self.config.get(
            "assess.enable_scan_response"
        )

    def is_assess_rule_disabled(self, rule_id):
        """
        Rules disabled in config override all disabled rules from TS per common config
        """
        app_settings = (
            self.direct_ts_app_settings
            if self.config.is_service_bypassed
            else self.application_settings
        )
        return (
            rule_id in self.disabled_assess_rules
            if self.disabled_assess_rules
            else rule_id in app_settings.disabled_assess_rules
        )

    def enabled_response_rules(self):
        return [
            rule
            for rule in self.assess_response_rules
            if not self.is_assess_rule_disabled(rule.name)
        ]

    def is_collect_stacktraces_all(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "ALL"

    def is_collect_stacktraces_some(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "SOME"

    def is_collect_stacktraces_none(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "NONE"

    def build_proxy_url(self):
        if self.proxy_url:
            return {self.proxy_scheme: self.proxy_url}

        return {}

    def update_sampling(self):
        server_sampling = (
            self.direct_teamserver_server_features.features.get("features")
            .get("assessment")
            .get("sampling")
            if self.config.is_service_bypassed
            else self.server_features.assess.sampling
        )

        sampling_lock_config = SamplingLockConfig()

        if not sampling_lock_config.lock_enable:
            self.config.put(
                "assess.sampling.enable",
                server_sampling.get("enable")
                if self.config.is_service_bypassed
                else server_sampling.enabled,
            )

        if not sampling_lock_config.lock_baseline:
            self.config.put(
                "assess.sampling.baseline",
                server_sampling.get("baseline")
                if self.config.is_service_bypassed
                else server_sampling.baseline,
            )

        if not sampling_lock_config.lock_frequency:
            self.config.put(
                "assess.sampling.request_frequency",
                server_sampling.get("frequency")
                if self.config.is_service_bypassed
                else server_sampling.request_frequency,
            )

        if not sampling_lock_config.lock_window:
            self.config.put(
                "assess.sampling.window_ms",
                server_sampling.get("window")
                if self.config.is_service_bypassed
                else server_sampling.window_ms,
            )

        logger.debug(
            "Updated sampling setting with TeamServer sampling settings: %s",
            server_sampling,
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import base64

import requests.models
from requests import put, post

import contrast
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.settings import Settings
from contrast.api import dtm_pb2
from contrast.utils.timer import now_ms
from contrast.utils.decorators import fail_loudly
from contrast.utils.object_utils import NOTIMPLEMENTED_MSG
from contrast.utils.service_util import sleep
from contrast.utils.string_utils import ensure_string
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

PYTHON = "Python"

EVENT_ACTION_LOOKUP = {
    0: "CREATION",
    1: "PROPAGATION",
    2: "TRIGGER",
    3: "TAG",
    4: "A2A",
    5: "A2P",
    6: "A2O",
    7: "A2R",
    8: "O2A",
    9: "O2P",
    10: "O2O",
    11: "O2R",
    12: "P2A",
    13: "P2P",
    14: "P2O",
    15: "P2R",
}
EVENT_TYPE_LOOKUP = {
    0: "METHOD",
    1: "PROPAGATION",
    2: "TAG",
}


SLEEP_TIME_SECS = 900


class BaseTsMessage:
    def __init__(self):
        self._sent_count = 0
        self.settings = Settings()

        self.base_url = f"{self.settings.api_url}/api/ng/"
        self.proxy = (
            self.settings.build_proxy_url() if self.settings.is_proxy_enabled else {}
        )

        self.server_name_b64 = _b64url_stripped(self.settings.get_server_name())
        self.server_path_b64 = _b64url_stripped(self.settings.get_server_path())
        self.server_type_b64 = _b64url_stripped(self.settings.get_server_type())
        auth_header = f"{self.settings.api_user_name}:{self.settings.api_service_key}"

        self.headers = {
            # the Authorization header must not have its padding stripped
            "Authorization": base64.urlsafe_b64encode(auth_header.encode()).decode(),
            "API-Key": self.settings.api_key,
            "Server-Name": self.server_name_b64,
            "Server-Path": self.server_path_b64,
            "Server-Type": self.server_type_b64,
            "X-Contrast-Agent": f"{PYTHON} {contrast.__version__}",
            "X-Contrast-Header-Encoding": "base64",
        }

        self.body = ""

    @property
    def class_name(self):
        return type(self).__name__.lstrip("_")

    @property
    def name(self) -> str:
        """
        Used for request audit filename
        """
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    @property
    def path(self) -> str:
        """
        URL path for teamserver; used for formatting as "/api/ng/{path}"
        """
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    @property
    def request_method(self) -> str:
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    @property
    def expected_response_codes(self):
        return [204]

    @property
    def sent_count(self):
        return self._sent_count

    def sent(self):
        self._sent_count += 1

    @fail_loudly("Failed to process TS response")
    def process_response(
        self, response: requests.models.Response, reporting_client
    ) -> None:
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    def process_response_code(self, response, reporting_client):
        """
        Return True if response code is expected response code
        """
        if not isinstance(response, requests.models.Response):
            return False

        logger.debug(
            "Received %s response code from Teamserver",
            response.status_code,
            direct_to_teamserver=1,
        )

        if response.status_code in (409, 410, 412, 502):
            # 409: app is archived, 502 app is locked in TS
            # 410: app is not registered. We could send App startup for for not we won't
            # 412: API key no longer valid. While spec may say to resend msg in 15mins,
            #  in reality the app server and agent should simply be restarted.
            DisableReaction.run(self.settings)
            return False

        if response.status_code in (401, 408):
            # 401: Access forbidden because credentials failed to authenticate.
            # 408: TS Could not create settings in time.

            if isinstance(self, (ServerStart, ApplicationStartup)):
                DisableReaction.run(self.settings)
                return False

            logger.debug(
                "Sleeping for 15 minutes",
                direct_to_teamserver=1,
            )

            sleep(SLEEP_TIME_SECS)

            reporting_client.retry_message(self)
            return False

        if response.status_code == 429:
            sleep_time = int(response.headers.get("Retry-After", SLEEP_TIME_SECS))

            logger.debug("Sleeping for %s seconds", sleep_time, direct_to_teamserver=1)

            sleep(sleep_time)

            reporting_client.retry_message(self)

        if response.status_code not in self.expected_response_codes:
            logger.debug(
                "Unexpected %s response from TS: %s",
                self.class_name,
                response.status_code,
                direct_to_teamserver=1,
            )
            return False

        return True


class BaseServerMsg(BaseTsMessage):
    @fail_loudly(f"Failed to process server settings response")
    def process_response(self, response, reporting_client):
        settings = Settings()
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()

        settings.apply_ts_server_settings(body)

        settings.process_ts_reactions(body)


class ServerActivity(BaseServerMsg):
    def __init__(self):
        super().__init__()

        self.body = {"lastUpdate": self.since_last_update}

    @property
    def name(self):
        return "activity-server"

    @property
    def path(self):
        return "activity/server"

    @property
    def request_method(self):
        return put

    @property
    def expected_response_codes(self):
        return [200, 304]

    @property
    def since_last_update(self):
        """
        Time in ms since server have been updated.
        If never updated, then it's been 0ms since then.
        """
        if self.settings.last_server_update_time_ms == 0:
            return 0
        return now_ms() - self.settings.last_server_update_time_ms

    @fail_loudly("Failed to process ServerActivity response")
    def process_response(self, response, reporting_client):
        settings = Settings()

        # TS will not send server settings unless lastUpdate is >= 5 mins (300_000 ms)
        # The response in those cases is 304.
        if (
            not self.process_response_code(response, reporting_client)
            or response.status_code == 304
        ):
            return

        body = response.json()

        settings.apply_ts_server_settings(body)

        settings.process_ts_reactions(body)


class ServerStart(BaseServerMsg):
    def __init__(self):
        super().__init__()

        self.body = {
            "environment": self.settings.config.get("server.environment"),
            "tags": self.settings.config.get("server.tags"),
            "version": contrast.__version__,
        }

    @property
    def name(self):
        return "agent-startup"

    @property
    def path(self):
        return "servers/"

    @property
    def expected_response_codes(self):
        return [200]

    @property
    def request_method(self):
        return put


class BaseTsAppMessage(BaseTsMessage):
    def __init__(self):
        super().__init__()

        # App language should only be encoded for url paths, not for headers.
        self.app_language_b64 = _b64url_stripped(PYTHON)
        self.app_name_b64 = _b64url_stripped(self.settings.app_name)

        self.headers.update(
            {
                "Application-Language": PYTHON,
                "Application-Name": self.app_name_b64,
                "Application-Path": _b64url_stripped(self.settings.app_path),
            }
        )

    @property
    def since_last_update(self):
        """
        Time in ms since app settings have been updated.
        If never updated, then it's been 0ms since then.
        """
        if self.settings.last_app_update_time_ms == 0:
            return 0
        return now_ms() - self.settings.last_app_update_time_ms


class Preflight(BaseTsAppMessage):
    def __init__(self, findings, http_request):
        super().__init__()

        self.findings = findings
        self.http_request = http_request

        self.body = {"messages": []}
        for idx, finding in enumerate(self.findings):
            message = {
                "appLanguage": PYTHON,
                "appName": self.settings.app_name,
                "appPath": self.settings.app_path,
                "appVersion": self.settings.app_version,
                "code": "TRACE",
                "data": finding.preflight,
                "key": idx,
            }
            self.body["messages"].append(message)

    @property
    def name(self):
        return "preflight"

    @property
    def path(self):
        return "preflight"

    @property
    def request_method(self):
        return put

    @property
    def expected_response_codes(self):
        return [200]

    @fail_loudly("Failed to process Preflight response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.text
        finding_idxs_to_report = self._parse_body(body)
        for finding_idx in finding_idxs_to_report:
            finding = self.findings[finding_idx]
            traces_msg = _Traces(finding, self.http_request)
            reporting_client.add_message(traces_msg)

    @staticmethod
    def _parse_body(body):
        """
        A preflight response body is a comma-separated list of finding indices that
        should be reported in a Traces message. Some elements of this list will have a
        *, meaning TS needs an AppCreate message before it will accept this finding. For
        now, we do not send findings with a *.

        TODO: PYT-2119 handle * preflight findings
        """
        indices = body.strip('"').split(",")
        return [int(index) for index in indices if index.isdigit()]


class _Traces(BaseTsAppMessage):
    def __init__(self, finding, request: dtm_pb2.HttpRequest):
        super().__init__()

        self.headers.update({"Report-Hash": finding.preflight.split(",")[-1]})

        self.body = self._build_body(finding, request)

    @property
    def name(self) -> str:
        return "traces"

    @property
    def path(self) -> str:
        return "traces"

    @property
    def request_method(self) -> str:
        return put

    @fail_loudly("Failed to process Traces response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)

    def _build_body(self, finding, request: dtm_pb2.HttpRequest):
        path, _, querystring = request.raw.partition("?")

        return {
            "created": (
                finding.events[-1].timestamp_ms if len(finding.events) > 0 else now_ms()
            ),
            "events": [event.to_json() for event in finding.events],
            "properties": finding.properties,
            "request": {
                "body": ensure_string(request.request_body_binary),
                # the WSGI environ supports only one value per request header. However
                # the server decides to handle multiple headers, we're guaranteed to
                # have only unique keys in request.request_headers (since we iterate
                # over webob's EnvironHeaders). Thus, each value list here is length-1.
                "headers": {k: [v] for k, v in request.request_headers.items()},
                "method": request.method,
                "parameters": {
                    h.key: list(h.values)
                    for h in request.normalized_request_params.values()
                },
                "port": request.receiver.port,
                "protocol": request.protocol,
                "queryString": querystring,
                "uri": path,
                "version": request.version,
            },
            "routes": [route.to_json_traces() for route in finding.routes],
            "ruleId": finding.rule_id,
            "session_id": self.settings.config.session_id,
            "tags": self.settings.config.assess_tags,
            "version": finding.version,
        }


class ObservedRoute(BaseTsAppMessage):
    def __init__(self, observed_route):
        # This message does not need "Application-Path" header but it doesn't hurt
        # either.
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

        self.body = {
            "signature": observed_route.signature,
            "verb": observed_route.verb,
            "url": observed_route.url,
            "sources": self._parse_sources(observed_route.sources),
        }

        session_id = self.settings.config.session_id
        if session_id:
            self.body.update({"session_id": session_id})

    @property
    def name(self):
        return "observed-route"

    @property
    def path(self):
        return (
            f"routes/{self.server_name_b64}/{self.server_path_b64}"
            f"/{self.server_type_b64}/{self.app_language_b64}/{self.app_name_b64}"
            "/observed"
        )

    @property
    def request_method(self):
        return post

    @fail_loudly("Failed to process ObservedRoute response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)

    def _parse_sources(self, sources):
        new_sources = []
        for source in sources:
            try:
                new_sources.append(dict(type=source.type, name=source.name))
            except Exception:
                logger.debug("Could not parse source %s", source)

        return new_sources


class ApplicationUpdate(BaseTsAppMessage):
    def __init__(self, libraries):
        super().__init__()

        self.headers.update({"Content-Type": "application/json"})

        # activity message sends "components" aka "architectures"
        # so we will not send the "components" field at this time.

        # field "timestamp" represents the amount of time that has passed
        # since the app settings were changed (not an actual timestamp).
        self.body = {
            "timestamp": self.since_last_update,
            "libraries": [
                lib.to_json(self.settings) for lib in libraries if lib.hash_code
            ],
        }

    @property
    def name(self):
        return "update-application"

    @property
    def path(self):
        return "update/application"

    @property
    def request_method(self):
        return put

    @fail_loudly("Failed to process ApplicationUpdate response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)


class ApplicationInventory(BaseTsAppMessage):
    def __init__(self, routes):
        # This message does not need "Application-Path" header but it doesn't hurt
        # either.
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

        self.body = {
            "routes": [route.to_json_inventory() for route in routes.values()],
        }

        session_id = self.settings.config.session_id
        if session_id:
            self.body.update({"session_id": session_id})

    @property
    def name(self):
        return "applications-inventory"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "inventory",
            ]
        )

    @property
    def request_method(self):
        return post

    @fail_loudly("Failed to process ApplicationInventory response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)


class ObservedLibraryUsage(BaseTsAppMessage):
    def __init__(self, observed_libraries):
        # This message does not need "Application-Path" header but it doesn't hurt
        # either.
        super().__init__()

        self.base_url = f"{self.settings.api_url}/agents/v1.0/"
        self.body = {"observations": []}

        for lib in observed_libraries:
            if len(lib.get("used_files")) > 0:
                message = {
                    "id": lib.get("hash_code"),
                    "names": list(lib.get("used_files")),
                }

                self.body["observations"].append(message)

    @property
    def name(self):
        return "applications-library-usage"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "library-usage",
            ]
        )

    @property
    def request_method(self):
        return post

    @fail_loudly("Failed to process ObservedLibraryUsage response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)


def _b64url_stripped(header_str):
    """
    For some headers, TS expects a value that
    - is base64 encoded using URL-safe characters
    - has any padding (= or ==) stripped

    This follows RFC-4648 - base64 with URL and filename safe alphabet
    """
    return base64.urlsafe_b64encode(header_str.encode()).rstrip(b"=").decode("utf-8")


class HeartBeat(BaseTsAppMessage):
    def __init__(self):

        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

    @property
    def name(self):
        return "applications-heartbeat"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "heartbeat",
            ]
        )

    @property
    def request_method(self):
        return post

    @fail_loudly("Failed to process HeartBeat response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)


class ApplicationStartup(BaseTsAppMessage):
    def __init__(self):
        super().__init__()

        self.body = {
            "instrumentation": {
                "protect": {"enable": self.settings.is_protect_enabled()}
            }
        }

        if self.settings.config.get_session_metadata():
            self.body.update(
                {"session_metadata": self.settings.config.get_session_metadata()}
            )

        if self.settings.config.app_metadata:
            self.body.update({"metadata": self.settings.config.app_metadata})

        if self.settings.config.app_code:
            self.body.update({"code": self.settings.config.app_code})

        if self.settings.config.session_id:
            self.body.update({"session_id": self.settings.config.session_id})

        if self.settings.config.app_tags:
            self.body.update({"tags": self.settings.config.app_tags})

        if self.settings.config.app_group:
            self.body.update({"group": self.settings.config.app_group})

    @property
    def name(self):
        return "applications-create"

    @property
    def path(self):
        return "applications/create"

    @property
    def expected_response_codes(self):
        return [200]

    @property
    def request_method(self):
        return put

    @fail_loudly("Failed to process ApplicationStartup response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()

        self.settings.apply_ts_app_settings(body)
        self.settings.process_ts_reactions(body)


class Activity(BaseTsAppMessage):
    def __init__(self, context):
        super().__init__()

        self.body = {"lastUpdate": self.since_last_update}

        self.body["inventory"] = {
            # Used by TeamServer to aggregate counts across a given time period, for Protect and attacker activity.
            "activityDuration": 0,
            "components": [],
        }
        if context.request.user_agent:
            self.body["inventory"]["browsers"] = [context.request.user_agent]
        if context.database_info:
            for db_info in context.database_info:
                self.body["inventory"]["components"].append(db_info)

        if context.attacks:
            self.body["defend"] = {"attackers": []}

            for attack in context.attacks:
                self.body["defend"]["attackers"].append(
                    {
                        "protectionRules": {
                            attack.rule_id: attack.to_json(
                                context.activity.http_request
                            )
                        },
                        "source": {
                            "ip": context.request.client_addr or "",
                            "xForwardedFor": context.request.headers.get(
                                "X-Forwarded-For"
                            )
                            or "",
                        },
                    }
                )

    @property
    def name(self):
        return "activity-application"

    @property
    def path(self):
        return "activity/application"

    @property
    def request_method(self):
        return put

    @property
    def expected_response_codes(self):
        return [200, 204]

    @fail_loudly("Failed to process Activity response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from enum import IntEnum

from contrast.extern import structlog as logging
from contrast.reporting.teamserver_messages import (
    EVENT_TYPE_LOOKUP,
    EVENT_ACTION_LOOKUP,
)

logger = logging.getLogger("contrast")


class TraceEvent(object):
    """
    Wrapper around api.dtm_pb2.TraceEvent
    """

    class EventType(IntEnum):
        TYPE_METHOD = 0
        TYPE_PROPAGATION = 1
        TYPE_TAG = 2

    class Action(IntEnum):
        CREATION = 0
        PROPAGATION = 1
        TRIGGER = 2
        TAG = 3
        A2A = 4
        A2P = 5
        A2O = 6
        A2R = 7
        O2A = 8
        O2P = 9
        O2O = 10
        O2R = 11
        P2A = 12
        P2P = 13
        P2O = 14
        P2R = 15

    def __init__(self):
        self.action = self.Action.CREATION
        self.type = self.EventType.TYPE_METHOD
        self.timestamp_ms = 0
        self.thread = ""
        self.signature = TraceEventSignature()  # contrast.api.dtm.TraceEventSignature
        self.field_name = ""
        self.context = ""
        self.code = ""
        self.object = TraceEventObject()  # contrast.api.dtm.TraceEventObject
        self.ret = TraceEventObject()  # contrast.api.dtm.TraceEventObject
        self.args = []  # contrast.api.dtm.TraceEventObject
        self.stack = []  # contrast.api.dtm.TraceStack
        self.event_sources = []  # contrast.api.dtm.TraceEventSource
        self.tags = []

        self.source = ""
        self.target = ""
        self.taint_ranges = []  # repeated contrast.api.dtm.TraceTaintRange

        self.object_id = 0
        self.parent_object_ids = []  # repeated contrast.api.dtm.ParentObjectId

    def to_json(self):
        return {
            "action": EVENT_ACTION_LOOKUP[self.action],
            "args": [
                {
                    # "hash": 0,  # not required
                    "tracked": arg.tracked,
                    "value": arg.value,
                }
                for arg in self.args
            ],
            # "code": "string",  # currently unused; maybe useful in the future
            "eventSources": [
                {
                    "sourceName": s.name,
                    "sourceType": s.type,
                }
                for s in self.event_sources
            ],
            "fieldName": self.field_name,
            "object": {
                # "hash": 0,  # not required
                "tracked": self.object.tracked,
                "value": self.object.value,
            },
            "objectId": self.object_id,
            "parentObjectIds": [{"id": p.id} for p in self.parent_object_ids],
            # properties not used for dataflow rules
            # "properties": [{"key": "string", "value": "string"}],
            "ret": {
                # "hash": 0,  # not required
                "tracked": self.ret.tracked,
                "value": self.ret.value,
            },
            "signature": {
                "argTypes": list(self.signature.arg_types),
                "className": self.signature.class_name,
                "constructor": self.signature.constructor,
                # not required and we don't save this to the DTM
                # "expressionType": "MEMBER_EXPRESSION",
                # "flags": 0,  # java only
                "methodName": self.signature.method_name,
                # not required and we don't save this to the DTM
                # "operator": "string",
                "returnType": self.signature.return_type,
                # "signature": "string",  # deprecated
                "voidMethod": self.signature.void_method,
            },
            "source": self.source,
            "stack": [
                {
                    "eval": s.eval,
                    "file": s.file_name,
                    "lineNumber": s.line_number,
                    "method": s.method_name,
                    "signature": s.signature,
                    "type": s.type,
                }
                for s in self.stack
            ],
            # "tags": "string",  # we don't save this to the DTM
            "taintRanges": [
                {
                    "tag": t.tag,
                    "range": t.range,
                }
                for t in self.taint_ranges
            ],
            "target": self.target,
            # "thread": "string",  # not required
            "time": self.timestamp_ms,
            "type": EVENT_TYPE_LOOKUP[self.type],
        }


class TraceEventObject(object):
    """
    Wrapper around api.dtm_pb2.TraceEventObject
    """

    def __init__(self):
        self.tracked = False
        self.value = ""
        self.ranges = []  # contrast.api.dtm.TraceTaintRange


class TraceEventSignature(object):
    """
    Wrapper around api.dtm_pb2.TraceEventSignature
    """

    def __init__(self):
        self.return_type = ""
        self.class_name = ""
        self.method_name = ""
        self.arg_types = []
        self.constructor = False
        self.void_method = False
        self.flags = 0


class ParentObjectId(object):
    """
    Wrapper around api.dtm_pb2.ParentObjectId
    """

    def __init__(self):
        self.id = ""


class TraceEventSource(object):
    """
    Wrapper around api.dtm_pb2.TraceEventSource
    """

    def __init__(self):
        self.type = ""
        self.name = ""


class TraceStack(object):
    """
    Wrapper around api.dtm_pb2.TraceStack
    """

    def __init__(self):
        self.signature = ""
        self.declaring_class = ""
        self.method_name = ""
        self.file_name = ""
        self.line_number = 0
        self.type = ""
        self.eval = ""


class TraceTaintRange(object):
    """
    Wrapper around api.dtm_pb2.TraceTaintRange
    """

    def __init__(self):
        self.tag = ""
        self.range = ""



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent import scope

from contrast.agent.request import Request
from contrast.agent.settings import Settings
from contrast.api.dtm_pb2 import Activity
from contrast.api.route_coverage import Route

from contrast.extern import structlog as logging
from contrast.utils.digest_utils import Digest
from contrast.utils.decorators import cached_property, fail_quietly
from contrast.utils.string_utils import ensure_binary, truncate

logger = logging.getLogger("contrast")


class RequestContext(object):
    def __init__(self, environ):
        scope.enter_contrast_scope()

        self.request = Request(environ)

        dtm = self.request.get_dtm()
        self.activity = Activity()
        self.activity.http_request.CopyFrom(dtm)

        # This contains input exclusions that should always apply regardless of rule on this request
        self.input_exclusions = None

        # This contains exclusions that only apply at trigger time based on the source event
        self.input_exclusions_trigger_time = None

        self.excluded_assess_rules = []
        self.excluded_protect_rules = []

        # For protect: store attacks made during a request to report at the end
        self.attacks = []

        # For assess: store findings made during a request to report at the end
        self.findings = []

        self.speedracer_input_analysis = None
        self.user_input_analysis = []
        self.response = None
        self.do_not_track = False

        # to be populated with a Route instance
        self.current_route = None

        self.observed_route = Route("", "", "")
        self.database_info = []
        self.source_count = 0
        self.propagation_count = 0

        self.max_sources_logged = False
        self.max_propagators_logged = False

        scope.exit_contrast_scope()

    @cached_property
    @fail_quietly("Unable to compute request hash")
    def hash(self):
        """
        Generates the hash checksum for the request. Converts the request method, uri,
        param names and content length to CRC checksum and returns string representation

        :return: str
        """
        hasher = Digest()

        hasher.update(self.request.method)
        hasher.update(self.request.get_normalized_uri())
        for key in self.request.params:
            hasher.update(key)

        hasher.update(self.request.content_length)
        return hasher.finish()

    def stop_source_creation(self, context, source_type, source_name):
        """
        Compare `source_count` to `max_context_source_events` config option
        :return: true if source_count within this request is equal to or
                 greater than configured threshold for source creation
        """
        if self.max_sources_logged:
            return True

        if context.input_exclusions and Settings().exclusions.evaluate_input_exclusions(
            context, source_type, source_name
        ):
            return True

        threshold_reached = self.source_count >= Settings().max_sources
        if threshold_reached:
            logger.warning(
                "Will not create more sources in this request. %s sources reached",
                self.source_count,
            )
            self.max_sources_logged = True

        return threshold_reached

    @property
    def stop_propagation(self):
        """
        Compare `source_count` to `max_propagation_events` config option
        :return: true if propagation_count within this request is equal to or
                 greater than configured threshold for propagation
        """
        if self.max_propagators_logged:
            return True

        threshold_reached = self.propagation_count >= Settings().max_propagation
        if threshold_reached:
            logger.warning(
                "Will not propagate any more in this request. %s propagations reached.",
                self.propagation_count,
            )
            self.max_propagators_logged = True

        return threshold_reached

    @property
    def propagate_assess(self):
        # TODO: PYT-644 move this property of out this class?
        return Settings().is_assess_enabled() and not scope.in_scope()

    def source_created(self):
        """
        Increase the running count of sources created in this request
        """
        self.source_count += 1

    def propagated(self):
        """
        Increase the running count of propagations created in this request
        """
        self.propagation_count += 1

    def extract_response_to_context(self, response):
        """
        Append response to request context and populate the HttpResponse DTM

        :response: Subclass of BaseResponseWrapper
        """
        self.response = response

        if not Settings().response_scanning_enabled:
            return

        self.activity.http_response.response_code = response.status_code

        # From the dtm for normalized_response_headers:
        #   Key is UPPERCASE_UNDERSCORE
        #
        #   Example: Content-Type: text/html; charset=utf-8
        #   "CONTENT_TYPE" => Content-Type,["text/html; charset=utf8"]
        for key, values in response.headers.dict_of_lists().items():
            normalized_key = key.upper().replace("-", "_")
            response_headers = self.activity.http_response.normalized_response_headers
            response_headers[normalized_key].key = key
            response_headers[normalized_key].values.extend(values)

        self.activity.http_response.response_body_binary = ensure_binary(
            response.body or ""
        )

    def truncate_request_body(self):
        # Request body is truncated to not overwhelm ContrastUI.
        raw_request_body = self.activity.http_request.request_body_binary
        self.activity.http_request.request_body_binary = truncate(
            raw_request_body, length=4096
        )
